// AP(r) Computer Science Marine Biology Simulation:
// The DrawableImageDisplay class based on FishImageDisplay
//
// This class is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation.
//
// This class is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

import java.awt.Color;
import java.awt.Component;
import java.awt.MediaTracker;
import java.awt.Image;
import java.awt.image.FilteredImageSource;
import java.awt.image.RGBImageFilter;
import java.awt.Graphics2D;
import java.awt.geom.Ellipse2D;
import javax.swing.ImageIcon;
import java.util.HashMap;

/**
 *  AP&reg; Computer Science Marine Biology Simulation:<br>
 *  A <code>DrawableImageDisplay</code> draws a drawable 
 *  object (an object of the <code>Drawable</code>
 *  interface using an
 *  image read from a file.  The image will be tinted to the object's color.
 *  The direction the original image is facing is supplied to the
 *  constructor, drawing is rotated as needed for the other directions.
 *
 *  <p>
 *  The <code>DrawableImageDisplay</code> 
 *
 * @author Barb Ericson 
 *  @author Julie Zelenski
 *  @version 1 August 2002
 **/
 
public class DrawableImageDisplay 
  extends AbstractDrawableDisplay
{
    private ImageIcon icon;
    private Image originalImage;
    private Direction originalDirection;
    private HashMap tintedVersions = new HashMap();
    
    /** Constructs an object that knows how to display a
     *  fish as an image. Looks for the named file first in
     *  the jar file, then in the current directory. If the named
     *  file is not found or the file is malformed, a colored
     *  circle will be substituted instead.
     *  @param imageFilename  name of file containing image
     *  @param dir            direction fish is facing in image
     **/
    public DrawableImageDisplay(String imageFilename, Direction dir)
    {    
        java.net.URL urlInJarFile = getClass().getResource(imageFilename);
        if (urlInJarFile != null) 
            icon = new ImageIcon(urlInJarFile);
        else 
        {
            String path = System.getProperty("user.dir") + java.io.File.separator + imageFilename;
            icon = new ImageIcon(path);
        } 
        originalDirection = dir;
        originalImage = icon.getImage();
    }
  
    /** Draws a unit-length fish facing North.
     *  This implementation draws a drawable object by tinting, scaling, and
     *  rotating the image provided to the constructor.
     *  @param   drawable      object we want to draw
     *  @param   comp       the component we're drawing on
     *  @param   g2         drawing surface
     **/
    public void draw(Drawable drawable, Component comp, Graphics2D g2)
    {
        if (icon.getImageLoadStatus() != MediaTracker.COMPLETE) 
        {
            // Image failed to load, so fall back to default display.
            g2.setPaint(drawable.color());
            g2.fill(new Ellipse2D.Double(-.4, -.4, .8, .8));
            return;
        }
    
        // Rotate drawing surface to compensate for the direction of the fish
        // in the image (in case it is not facing North, as assumed). 
        if (!originalDirection.equals(Direction.NORTH))
        {
            int rotationInDegrees = originalDirection.inDegrees();
            if (rotationInDegrees >= 180) g2.scale(-1, 1); // flip image side-to-side
            g2.rotate(- Math.toRadians(rotationInDegrees));
        }
    
        // Scale to shrink or enlarge the image to fit the size 1x1 cell.
        g2.scale(1.0/icon.getIconWidth(), 1.0/icon.getIconHeight());

       // Compose image with fish color using an image filter.
        Color drawableColor = drawable.color();
        Image tinted = (Image)tintedVersions.get(drawableColor);
        if (tinted == null)  // not cached, need new filter for color
        {
            FilteredImageSource src = new FilteredImageSource(originalImage.getSource(), 
                                                  new TintFilter(drawableColor));
            tinted = comp.createImage(src);
            // Cache tinted image in map by color, we're likely to need it again.
            tintedVersions.put(drawableColor, tinted);
        }
        
        icon.setImage(tinted);
        icon.paintIcon(comp, g2, -icon.getIconWidth()/2, -icon.getIconHeight()/2);    
    }

    /** An image filter class that tints colors based on the tint provided
     *  to the constructor (the color of a fish).
     **/
    private static class TintFilter extends RGBImageFilter 
    {
        private int tintR, tintG, tintB;
        
        /** Constructs an image filter for tinting colors in an image. **/
        public TintFilter(Color color)
        {
            canFilterIndexColorModel = true;
            int rgb = color.getRGB();
            tintR = (rgb >> 16) & 0xff;
            tintG = (rgb >> 8) & 0xff;
            tintB =  rgb & 0xff;
        }
                
        public int filterRGB(int x, int y, int argb)
        {
            // Separate pixel into its RGB coomponents.
            int alpha = (argb >> 24) & 0xff;
            int red =   (argb >> 16) & 0xff;
            int green = (argb >> 8) & 0xff;
            int blue =   argb & 0xff;
            
            // Use NTSC/PAL algorithm to convert RGB to grayscale.
            int lum = (int) (0.2989 * red + 0.5866 * green + 0.1144 * blue);
            
            // Interpolate along spectrum black->white with tint at midpoint
            double scale = Math.abs((lum - 128)/128.0); // absolute distance from midpt
            int edge = lum < 128 ? 0 : 255; // going towards white or black?
            red =   tintR + (int)((edge - tintR) * scale); // scale from midpt to edge
            green = tintG + (int)((edge - tintG) * scale);
            blue =  tintB + (int)((edge - tintB) * scale);
            return (alpha << 24) | (red << 16) | (green << 8) | blue;
        }

    }

}
