// AP(r) Computer Science Marine Biology Simulation:
// The MBSDisplay class is copyright(c) 2002 College Entrance
// Examination Board (www.collegeboard.com).
//
// This class is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation.
//
// This class is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

import java.awt.*;
import javax.swing.*;
import java.awt.event.MouseEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.geom.AffineTransform;

/**
 *  AP&reg; Computer Science Marine Biology Simulation:<br>
 *  A <code>MBSDisplay</code> is a panel containing a graphical display of
 *  the environment of fish.
 *
 *  <p>
 *  The <code>MBSDisplay</code> class is
 *  copyright&copy; 2002 College Entrance Examination Board
 *  (www.collegeboard.com).
 *
 *  @author Julie Zelenski
 *  @version 1 August 2002
 **/
 
public class MBSDisplay extends JPanel implements EnvDisplay, Scrollable, PseudoInfiniteViewport.Pannable
{
    // Class constants
    private static final int MIN_CELL_SIZE = 8, DEFAULT_CELL_SIZE = 16;
    private static final Color OCEAN_BLUE = new Color(75, 75, 255);
    
    // Encapsulated data used to monitor/display the environment
    private Environment theEnv;
    private int numRows, numCols, originRow, originCol;
    private int cellSize;
    private boolean toolTipsEnabled;


    /** Construct a new MBSDisplay object for the given environment.
     *  @param env the Environment to display
     **/
    public MBSDisplay(Environment env)
    {
        setToolTipsEnabled(true);   
        addComponentListener(new ComponentAdapter() {
            public void componentResized(ComponentEvent e)
                { if (getEnclosingViewport() == null) recalculateCellSize(MIN_CELL_SIZE); }});
        if (env != null) setEnv(env);
    }
    
    /** Construct a new MBSDisplay object with no environment.
     *  The view will be empty.
     **/
    public MBSDisplay()
    {
        this(null);
    }
    

    /** Show the environment.
     *  This is the one required method in the EnvDisplay interface.
     *  Invoking the <code>repaint</code> method is the standard way to ask a
     *  Swing component to redraw itself. This eventually turns into a call
     *  back to our version of the the standard <code>paintComponent</code> method 
     *  where we do the actual drawing.
     **/
    public void showEnv()
    {
        repaint();
    }

    /** Paint this component.
     *  @param g the Graphics object to use to render this component 
     **/
    public void paintComponent(Graphics g)
    {
        Graphics2D g2 = (Graphics2D)g;
        
        super.paintComponent(g2);
        if (theEnv == null) return;
        
        Insets insets = getInsets();
        g2.setColor(OCEAN_BLUE);  // fill blue background
        g2.fillRect(insets.left, insets.top, numCols*cellSize, numRows*cellSize);
        
        drawGridlines(g2);
    
        Locatable[] envObjects = theEnv.allObjects();
        for (int k = 0; k < envObjects.length; k++) 
            drawLocatable(g2, envObjects[k]);
   }
    
    /** Draw one environment object. First verify that the object
     *  is actually visible before any drawing, set up the clip
     *  appropriately and use the DisplayMap to determine
     *  which object to call upon to render this particular Locatable.
     *  Note that we save and restore the graphics transform to
     *  restore back to normalcy no matter what the renderer did to
     *  to the coordinate system. 
     *  @param g2 the Graphics2D object to use to render 
     *  @param obj the Locatable object to draw
     **/
    private void drawLocatable(Graphics2D g2, Locatable obj)
    {
        Rectangle cellToDraw = new Rectangle(colToXCoord(obj.location().col()) + 1, 
                                             rowToYCoord(obj.location().row()) + 1, 
                                             cellSize - 1, cellSize - 1);
    
        // Only draw if the object is visible within the current clipping region.
        if (cellToDraw.intersects(g2.getClip().getBounds())) 
        { 
            AffineTransform savedTransform = g2.getTransform(); // save
            
            // Get the drawing object to display this environment object.
            LocatableDisplay displayObj = DisplayMap.findDisplayFor(obj);
            displayObj.draw(obj, this, g2, cellToDraw);
            
            g2.setTransform(savedTransform); // restore coordinate system
        }
    }
    
    
    /** Draw the gridlines for the environment. We only draw the portion
     *  of the lines that intersect the current clipping bounds.
     *  @param g2 the Graphics2 object to use to render 
     **/
    private void drawGridlines(Graphics2D g2)
    {
        Rectangle curClip = g2.getClip().getBounds();
        int top = getInsets().top, left = getInsets().left;
        
        int miny = Math.max(0, (curClip.y - top)/cellSize)*cellSize + top;
        int minx = Math.max(0, (curClip.x - left)/cellSize)*cellSize + left;
        int maxy = Math.min(numRows, (curClip.y + curClip.height - top + cellSize - 1)/cellSize)*cellSize + top;
        int maxx = Math.min(numCols, (curClip.x + curClip.width - left + cellSize - 1)/cellSize)*cellSize + left;

        g2.setColor(Color.black);
        g2.setStroke(new BasicStroke());
        for (int y = miny; y <= maxy; y += cellSize) // draw horizontal lines
            g2.drawLine(minx, y, maxx, y);
    
        for (int x = minx; x <= maxx; x += cellSize) // draw vertical lines
            g2.drawLine(x, miny, x, maxy);
    }
    
    /** Enables/disables showing of tooltip giving information about
     *  the environment object beneath the mouse.
     *  @param flag whether to enable/disable tool tips
     **/
    public void setToolTipsEnabled(boolean flag)
    {
        if (flag)
            ToolTipManager.sharedInstance().registerComponent(this);  
        else
            ToolTipManager.sharedInstance().unregisterComponent(this);
 toolTipsEnabled = flag;
    }

    /** Sets the Environment being displayed.  Reset the cellSize to
     *  be the largest that fits the entire environment in the current 
     *  visible area (use default if environment is too large).
     *  @param newEnv the Environment to display
     **/
    public void setEnv(Environment newEnv)
    {
        JViewport vp = getEnclosingViewport(); // before changing, reset scroll/pan position
        if (vp != null) vp.setViewPosition(new Point(0, 0));

        theEnv = newEnv;
        originRow = originCol = 0;
        
        if (theEnv.numRows() == -1 && theEnv.numCols() == -1)
        {         
            numRows = numCols = 2000; // This determines the "virtual" size of the pan world
        }
        else
        {
            numRows = theEnv.numRows();
            numCols = theEnv.numCols();
        }
        recalculateCellSize(DEFAULT_CELL_SIZE);
    }
    
    // private helpers to caculate extra width/height needs for borders/insets.
    private int extraWidth() 
    { 
        return getInsets().left + getInsets().right + 1;
    }
    
    private int extraHeight() 
    { 
        return getInsets().top + getInsets().left + 1; 
    }
    
    /** Returns the desired size of the display, for use by layout manager.
     *  @return preferred size
     **/
    public Dimension getPreferredSize()
    {
       return new Dimension(numCols*cellSize + extraWidth(), numRows*cellSize + extraHeight());
    }
    
    /** Returns the minimum size of the display, for use by layout manager.
     *  @return minimum size
     **/
    public Dimension getMinimumSize()
    {
        return new Dimension(numCols*MIN_CELL_SIZE + extraWidth(), numRows*MIN_CELL_SIZE + extraHeight());
    }
    
    /** Zooms in the display by doubling the current cell size.
     **/
    public void zoomIn()
    {
        cellSize *= 2;
        revalidate();
    }
    
    /** Zooms out the display by halving the current cell size.
     **/
    public void zoomOut()
    {
        cellSize = Math.max(cellSize/2, MIN_CELL_SIZE);
        revalidate();
    }

    /** Pans the display back to the origin, so that 0, 0 is at the
     *  the upper left of the visible viewport.
     **/
    public void recenterOnOrigin()
    {
        originRow = 0;
        originCol = 0;
        repaint();
        JViewport vp = getEnclosingViewport();
        if (vp != null)
        {
            if (!isPannableUnbounded() || !(vp instanceof PseudoInfiniteViewport))
                vp.setViewPosition(new Point(0, 0));
            else
                ((PseudoInfiniteViewport)vp).showOriginTip();
        }
    }
    
    /** Given a Point determine which environment location (if any) is under the mouse.
     *  This method is used by the GUI when creating Actionable Objects by clicking on cells in the display.
     *  @param p the Point in question (in display's coordinate system)
     *  @return the environment Location beneath the event, or null if no cell is beneath the mouse
     **/
    public Location locationForPoint(Point p)
    {
        Location loc = new Location(yCoordToRow(p.y), xCoordToCol(p.x));
        return (theEnv != null && theEnv.isValid(loc)) ? loc : null;
    }

    // private helpers to convert between (x,y) and (row,col)
    private int xCoordToCol(int xCoord)
    {
        return (xCoord - getInsets().left)/cellSize + originCol;
    }
    private int yCoordToRow(int yCoord)
    {
        return (yCoord - getInsets().top)/cellSize + originRow;
    }
    private int colToXCoord(int col)
    {
        return (col - originCol)*cellSize + getInsets().left;
    }
    private int rowToYCoord(int row)
    {
        return (row - originRow)*cellSize + getInsets().top;
    }
    
    /** Given a MouseEvent, determine what text to place in the floating tool tip when the
     *  the mouse hovers over this location.  If the mouse is over a valid environment cell.
     *  we provide some information about the cell and its contents.  This method is 
     *  automatically called on mouse-moved events since we register for tool tips.
     *  @param evt the MouseEvent in question 
     *  @return the tool tip string for this location
     **/   
    public String getToolTipText(MouseEvent evt)
    {
        Location loc = locationForPoint(evt.getPoint());
        if (!toolTipsEnabled || loc == null)
            return null;
        AbstractDrawable a = (AbstractDrawable)theEnv.objectAt(loc);
        if (a != null)
            return a.getClass().getName() + " id #" + a.id()
                    + " at " + loc + " facing " + a.direction();
        else
            return loc + " is empty";
    }
    
    
        
    /** Calculate the cell size to use given the current viewable region and the
     *  the number of rows and columns in the environment.  We use the largest
     *  cellSize that will fit in the viewable region, bounded to be at least
     *  the parameter minSize.
     **/   
    private void recalculateCellSize(int minSize)
    {
        if (numRows == 0 || numCols == 0) 
        {
            cellSize = 0;
        } 
        else 
        {
            JViewport vp = getEnclosingViewport();
            Dimension viewableSize = (vp != null) ? vp.getSize() : getSize();
            cellSize = Math.min((viewableSize.height - extraHeight())/numRows, (viewableSize.width - extraWidth())/numCols);
            cellSize = Math.max(cellSize, minSize);
        }
        revalidate();
    }
    
    // helper to get our parent viewport, if we are in one.
    private JViewport getEnclosingViewport()
    {
 Component parent = getParent();
 return (parent instanceof JViewport) ? (JViewport)parent : null;
    }

   
    // MBSDisplay implements the Scrollable interface to get nicer behavior in a JScrollPane
    // The 5 methods below are the methods in that interface
    
    public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction) 
    { 
        return cellSize;
    }
    
    public int getScrollableBlockIncrement(Rectangle visibleRect, int orientation, int direction) 
    { 
        if (orientation == SwingConstants.VERTICAL)
            return (int)(visibleRect.height * .9);
        else
            return (int)(visibleRect.width * .9);
    }
    
    public boolean getScrollableTracksViewportWidth() 
    {  
        return false;
    }
    
    public boolean getScrollableTracksViewportHeight() 
    { 
         return false;
    }
    
    public Dimension getPreferredScrollableViewportSize() 
    {
        return new Dimension(420, 420);
    }
    
    // MBSDisplay implements the PseudoInfiniteViewport.Pannable interface to 
    // play nicely with the pan behavior for unbounded view.
    // The 3 methods below are the methods in that interface.

    public void panBy(int hDelta, int vDelta)
    {
        originCol += hDelta/cellSize;
        originRow += vDelta/cellSize;
        repaint();
    }
    
    public boolean isPannableUnbounded() 
    { 
        return (theEnv != null && theEnv.numRows() == -1); 
    }
    
    public String getPannableTipText()
    {
        Point upperLeft = new Point(0, 0);
        JViewport vp = getEnclosingViewport();
        if (!isPannableUnbounded() && vp != null) upperLeft = vp.getViewPosition();
        Location loc = locationForPoint(upperLeft);
        return (loc != null ? loc.toString() : null);
    }

}
